use std::collections::HashSet;

use aoc_runner_derive::{aoc, aoc_generator};
use itertools::Itertools;

#[derive(Clone, Copy, Hash, PartialEq, Eq)]
struct Item(char);

impl Item {
    fn new(c: char) -> Option<Self> {
        c.is_ascii_alphabetic().then_some(Self(c))
    }

    fn priority(self) -> u32 {
        match self.0 {
            'a'..='z' => self.0 as u8 - b'a' + 1,
            'A'..='Z' => self.0 as u8 - b'A' + 27,
            _ => unreachable!(),
        }
        .into()
    }
}

#[aoc_generator(day3)]
fn generator(s: &str) -> Vec<(HashSet<Item>, HashSet<Item>)> {
    s.lines()
        .map(|l| {
            let (a, b) = l.split_at(l.len() / 2);
            let a = a.chars().map(|v| Item::new(v).unwrap()).collect();
            let b = b.chars().map(|v| Item::new(v).unwrap()).collect();
            (a, b)
        })
        .collect()
}

#[aoc(day3, part1)]
fn part1(input: &[(HashSet<Item>, HashSet<Item>)]) -> u32 {
    input
        .iter()
        .map(|(a, b)| a.intersection(b).next().unwrap().priority())
        .sum()
}

#[aoc(day3, part2)]
fn part2(input: &[(HashSet<Item>, HashSet<Item>)]) -> u32 {
    input
        .iter()
        .tuples()
        .map(|(a, b, c)| {
            let a = a.0.union(&a.1).copied().collect::<HashSet<_>>();
            let b = b.0.union(&b.1).copied().collect::<HashSet<_>>();
            let c = c.0.union(&c.1).copied().collect::<HashSet<_>>();

            let v = a;
            let v = v.intersection(&b).copied().collect::<HashSet<_>>();
            let v = v.intersection(&c).copied().next().unwrap();
            v.priority()
        })
        .sum()
}
